;;; Automates cellulaires
;;; Programmation peu efficace, favorisant simplicit et lisibilit
;;; Implmentation en listes plutt qu'en vecteurs : moins efficace
;++ Cellular Automata
;++ Low efficiency programming, in favor of simplicity and legibility
;++ Implementation with lists rather than vectors: less efficient

;;   Cette fonction traite une squence initiale, en lui faisant subir une
;; transformation selon un ensemble de rgles pr-dfini. Chaque triplet de
;; "cellules" adjacentes de la squence initiale dtermine, selon la rgle
;; correspondante, le nouvel tat de la "cellule centrale" du triplet.
;+   This function transforms an initial sequence by applying a set of
;+ preestablished rules. Each triplet of adjacent "cells" of the initial
;+ sequence determines, according to the corresponding rule, the new state of
;+ the "center cell" of the triplet.
;-    (otomatesel <seq-presente> <regles>)
;; Des exemples de <regles> possibles sont dfinis ci-dessous.
;;   Exemple, automate  trois tats :
;+ Examples of possible <regles> rules are defined below.
;+   Example, three states automata:
;-    (otomatesel '(a b a c a b a) postulainit02) => (C A C B B A A)
;; Ritration, automate  deux tats :
;+ Reiteration, two states automata:
;-    (setq seqc '(a b a b b a b b a b b b a a))
;-    (pasapas (print (setq seqc (otomatesel seqc postulainit01))))

(defun otomatesel (seq-presente regles
                   &aux triplet laregle lacellule seq-nouvelle)
  ;premier triplet (cas particulier)
  (setq triplet (append (last seq-presente) (firstn2 seq-presente 2)))
  (setq laregle (member triplet regles :test #'equal))
  (setq lacellule (cadr laregle))
  (setq seq-nouvelle (list lacellule))
  ;tous autres triplets "normaux"
  (dotimes (compteur (- (list-length seq-presente) 2))
    ;(dbg seq-nouvelle)
    (setq triplet (firstn2 (nthcdr compteur seq-presente) 3))
    ;(dbg triplet)
    (setq laregle (member triplet regles :test #'equal))
    ;(dbg laregle)
    (setq lacellule (cadr laregle))
    ;(dbg lacellule)
    (setq seq-nouvelle (append seq-nouvelle (list lacellule))))
  ;dernier triplet (cas particulier)
  (setq triplet (append (last seq-presente 2) (firstn2 seq-presente 1)))
  (setq laregle (member triplet regles :test #'equal))
  (setq lacellule (cadr laregle))
  (setq seq-nouvelle (append seq-nouvelle (list lacellule))))

;;   Cette fonction reproduit la prcdente otomatesel autant de fois qu'on le
;; souhaite selon <nbrepas>. Exemple,  trois tats :
;+   This function repeats the preceding otomatesel as many times as desired,
;+ according to <nbrepas>. Example, with three states:
;-    (automatecell '(a b a c a b a) postulainit01 3 :gepresst nil)
;-        => (A B A C A B A)
;-           (C B C A B B A)
;-           (A C A B A A C)
;-           (C A B B A B A)
;-           OK

(defun automatecell (seq-depart regles nbrepas &key gepresst &aux pas+1)
  (if gepresst (tight-line seq-depart) (format t "~&~A~%" seq-depart))
  (dotimes (pas nbrepas 'ok)
    (setq pas+1 (otomatesel seq-depart regles))
    (if gepresst (tight-line pas+1) (format t "~&~A~%" pas+1))
    (setq seq-depart pas+1)))

;+   To print the list state as a tight sequence of single characters without
;+ white spaces.

(defun tight-line (list &aux tmp)
  (format t "~&")
  (dotimes (n (list-length list))
    (format t "~A" (if (eq '- (setq tmp (nth n list))) #\  tmp)))
  (format t "~%"))

;;   Pour ces fonctions, il est ncessaire d'avoir dtermin au pralable des
;; rgles de transformation entre les triplets de la squence initiale et le
;; nouvel tat de leur "cellule centrale". Ces ensembles de rgles pr-tablies
;; sont ici nomms "postulats initiaux".
;;   Dans le cas d'automates cellulaires  trois tats, chaque postulat consiste
;; donc en un ensemble de <3 tats possibles> ** <3 cellules> = 27 rgles.
;+   For these function one must have previously determined transformation rules
;+ between the triplets of the initial sequence and the new state of their
;+ "central cell. These sets of preestablished rules are here named "postulats
;+ initiaux".
;+   In the case of three states cellular automata, each "postulat" thus
;+ consists in a set of <3 possible states> ** <3 cells> = 27 rules.

(defparameter postulainit01 '(
    (a a a) a
    (a a b) c
    (a a c) b
    (a b a) b
    (a b b) a
    (a b c) c
    (a c a) a
    (a c b) a
    (a c c) c
    (b a a) a
    (b a b) a
    (b a c) c
    (b b a) a
    (b b b) b
    (b b c) a
    (b c a) a
    (b c b) b
    (b c c) a
    (c a a) a
    (c a b) b
    (c a c) c
    (c b a) a
    (c b b) b
    (c b c) c
    (c c a) a
    (c c b) b
    (c c c) c
))

;;   Une autre proposition de postulat. Ce choix pralable est arbitraire.
;+   Another possible set of rules. This initial choice is arbitrary.

(defparameter postulainit02 '(
    (a a a) a
    (a a b) c
    (a a c) b
    (a b a) a
    (a b b) a
    (a b c) c
    (a c a) b
    (a c b) a
    (a c c) c
    (b a a) a
    (b a b) b
    (b a c) c
    (b b a) c
    (b b b) b
    (b b c) a
    (b c a) a
    (b c b) b
    (b c c) b
    (c a a) a
    (c a b) b
    (c a c) c
    (c b a) a
    (c b b) c
    (c b c) c
    (c c a) a
    (c c b) c
    (c c c) c
))

;;   Exemple de postulat pour un automate  deux tats. Correspond  la rgle
;; dite "110" dans l'implmentation traditionnelle des automates cellulaires 
;; deux tats de Stephen Wolfram.
;;   Dans le cas d'automates cellulaires  deux tats, chaque postulat consiste
;; en un ensemble de <2 tats possibles> ** <3 cellules> = 8 rgles.
;+   "Postulat" example for a two states automata. Corresponds to so called rule
;+ "110" in Stephen Wolfram's traditional implementation of two states automata.
;+   In the case of two states cellular automata, each "postulat" consists in a
;+ set of <2 possible states> ** <3 cells> = 8 rules.

;+ Classical Wolfram case, rule "101".
;+ N.B. Use :gepresst t for an improved display
;-    (automatecell 
;-      '(- - - X - X X X - X X - - - X X - - X - X - X - - - - X X X - X - - X X X X X - X X - X - X X X X - - X X X - - - X - X - - - X - X - - X - X X - - X - - - X X X X - X - - X X - - - - - X X X X - - X X - - X - - X X - - - - - X X - - - - X X X X X - - - X X X X - X X X - - X X - - - - X X X X - X - - - - X X - - X - - - X X X - - - - X - - - X X - X - X - -)
;-      postul01-2 180 :gepresst nil)

(defparameter postul01-2 '(                                      ;Rule 101
    (- - -) X
    (- - X) -
    (- X -) X
    (- X X) -
    (X - -) -
    (X - X) X
    (X X -) X
    (X X X) -
))

;+ Classical Wolfram case, rule "54".
;+ N.B. Use :gepresst t for an improved display
;-    (automatecell 
;-     '(- - X - - - X X - - X X X X X X - X X - - - - - - X X - X - - - X - X X - - - X - - - - X - X - X X - - - X - - X - - - - - X - - X X X - X X X X X X - X X X X X X - - X X - - - - X - - X X - X X - - X - - - X - X - - - - - X - X X - - - - - X - X X - - - X - X X X X X - - - - - - - X X - X X X X - X X - - X X - - X X X - X X X X - - - - X - X - X X - - X -)
;-     postu54-2 180 :gepresst nil)

(defparameter postu54-2 '(                                      ;Rule 54
    (- - -) -
    (- - X) X
    (- X -) X
    (- X X) -
    (X - -) X
    (X - X) X
    (X X -) -
    (X X X) -
))

;+   Used by function otomatesel above. Cf. ped.08.

(defun firstn2 (li &optional (n 1))
"
;    firstn2 <liste> &optional (<n> 1) => <n> premier(s) element(s) de <liste>
"
  (butlast li (- (list-length li) n)))
