;;;=============================================================================
;;; 18.2 - Annexe C --> FILTRES LABORS (pattern matching) : liaison d'lments
;;;                     d'assertions    des  variables  de  filtres,  calcul de
;;;                     contraintes et d'lments de filtres
;;;
;;;               a. -> DFINITION SYNTAXIQUE DES FILTRES, MACRO-CARACTRES
;;;=============================================================================

;;   Les caractres spciaux suivants sont dfinis pour spcifier des lments
;; de filtres quivalents  certains nombres d'lments d'assertions -- cf.
;; Chap. 18.1 :
;;
;;    _         := un et un seul lment quelconque                   _ := [1]
;;    $         := un ou plusieurs lment(s) quelconque(s)           $ := [1,n[
;;    %         := zro ou un lment quelconque                      % := [0,1]
;;    ~         := zro, un ou plusieurs lment(s) quelconque(s)     ~ := [0,n[

;;   Le caractre spcial ?, suivi de l'un des prcdents, permet de dcrire de
;; vritables listes d'alternatives de dfinition d'un filtre, constitues soit
;; d'lments constants pr-dfinis, soit d'autres caractres spciaux, etc. En
;; particulier, les deux caractres spciaux juxtaposs ?_ permettent donc en
;; outre de spcifier, parmi les alternatives, des squences de plus d'un
;; lment de filtre.

;;   Le caractre spcial ^ permet de spcifier, en tant qu'lment de filtre,
;; une forme  valuer ultrieurement en tenant compte de l'lment
;; correspondant de l'assertion en cours de confrontation, lui-mme reprsent
;; par le symbole &. On peut ainsi inclure dans un filtre des contraintes
;; varies par rapport aux lments d'assertions, et nuancer librement les
;; conditions de succs ou d'chec du filtrage, sans se limiter soit  la
;; stricte galit, soit au simple dnombrement d'lments quelconques, qui sont
;; rendus possibles par les quatre caractres spciaux du premier ensemble
;; rduit { _ $ % ~ } ci-dessus.

;;   La syntaxe permet de crer des variables de filtre, qui sont lies, au
;; cours d'une opration de filtrage, aux lments appropris de l'assertion 
;; laquelle est confront le filtre.  l'issue du filtrage, ces variables de
;; filtre permettent de rcuprer, au besoin, les parties significatives de
;; l'assertion, en plus du simple constat de russite ou d'chec global. On peut
;; en outre examiner, aussi bien au cours du filtrage qu' posteriori, les
;; lments d'assertion qui sont accepts par le filtre -- soit  titre de
;; tentatives provisoires, soit dfinitivement. Ceci permet, d'une part,
;; d'auto-adapter le filtre lui-mme  l'assertion en cours de confrontation, ou
;; d'autre part, de tirer des conclusions pertinentes du succs ou de l'chec du
;; filtrage global, qui peuvent  leur tour donner matire  d'ventuelles
;; actions ultrieures.

;;   Le caractre spcial @ permet d'insrer dans le filtre un ou plusieurs
;; lment(s) calcul(s)  partir de l'valuation d'une forme, qui sera
;; effectue dans le contexte des variables du filtre en cours de confrontation.
;; On peut ainsi dfinir un(des) lment(s) de filtre par rapport  un(des)
;; lment(s) de l'assertion dj filtr(s) par un(des) lment(s) prcdent(s)
;; du mme filtre. Le filtre peut donc s'auto-adapter, au cours de sa
;; confrontation avec elle, aux caractristiques de l'assertion qu'il examine.

;;; Dfinitions syntaxiques

;;   Dfinitions syntaxiques et prparation effective des filtres,  l'aide de
;; l'ensemble des 7 caractres spciaux dcrits ci-dessus :
;;    { _ $ % ~ ? ^ @ },
;; traits en tant que macro-caractres.
;;   Le symbole &, dans le contexte du macro-caractre ^, sera en outre
;; interprt selon l'usage dcrit ci-dessus.
;;
;; a. Mta-dfinitions des lments syntaxiques
;;
;;         x  := un caractre spcial parmi { _  $ % ~ }
;;         Z  := nom de variable de filtre quelconque (symbole)
;;        (L) := liste quelconque, pouvant comporter d'autres occurrences de
;;               macro-caractres, etc.
;;        (C) := forme dfinissant une contrainte de filtrage, autre que
;;               l'galit, dans laquelle le symbole & tient lieu de l'lment
;;               d'assertion en cours de filtrage
;;        (F) := forme quelconque, dont l'valuation, dans le contexte des
;;               variables de filtre, est  insrer dans le filtre
;;
;; b. Rgles d'expansion des macro-caractres
;;
;;        x      -> (x)
;;        xZ     -> (x . Z)
;;        ?x(L)  -> ("?x" nil L)
;;        ?xZ(L) -> ("?x"  Z  L)
;;        ^(C)   -> (^ nil C)
;;        ^Z(C)  -> (^  Z  C)
;;        @Z     -> (@ . Z)
;;        @(F)   -> (@ F)

;;; Utilitaire pour les macro-caractres ci-dessous

;;   Cration d'un tableau de hachage spcifiant le type syntaxique des 128
;; caractres ASCII, et reconstitution de l'qauivalent de l'appel des
;; primitives Le_Lisp (typech (peekch)) -- cf. [<Manuel de rfrence Le_Lisp
;; 15.22>, INRIA, 1989 : 6-15].
;;   Les donnes proprement dites -- les types des caractres -- sont reportes
;; ci-dessous, en fin de fichier.

(defparameter common-lisp::typech (make-hash-table :size 128))

(defmacro typechpeek ()
  '(gethash (peek-char nil flux) common-lisp::typech))

;;; Dfinitions des macro-caractres et des fonctions correspondantes

(defun sousligne (flux caractere) caractere (cond                             ;_
    ((member (typechpeek) '(sep bcom ecom lpar rpar))
     '(|_|))
    ((not (eq 'pname (typechpeek)))
     (let ((car (read-char flux)))
       (error "macro-caractere \"_\" ne peut etre suivi de \"~A\"." car)))
    (t (cons '|_| (read flux)))))
(set-macro-character #\_ #'sousligne)

(defun piastre (flux caractere) caractere (cond                               ;$
    ((member (typechpeek) '(sep bcom ecom lpar rpar))
     '(|$|))
    ((not (eq 'pname (typechpeek)))
     (let ((car (read-char flux)))
       (error "macro-caractere \"$\" ne peut etre suivi de \"~A\"." car)))
    (t (cons '|$| (read flux)))))
(set-macro-character #\$ #'piastre)

(defun pourcent (flux caractere) caractere (cond                              ;%
    ((member (typechpeek) '(sep bcom ecom lpar rpar))
     '(|%|))
    ((not (eq 'pname (typechpeek)))
     (let ((car (read-char flux)))
       (error "macro-caractere \"%\" ne peut etre suivi de \"~A\"." car)))
    (t (cons '|%| (read flux)))))
(set-macro-character #\% #'pourcent)

(defun tilde (flux caractere) caractere (cond                                 ;~
    ((member (typechpeek) '(sep bcom ecom lpar rpar))
     '(|~|))
    ((not (eq 'pname (typechpeek)))
     (let ((car (read-char flux)))
       (error "macro-caractere \"~~\" ne peut etre suivi de \"~A\"." car)))
    (t (cons '|~| (read flux)))))
(set-macro-character #\~ #'tilde)

(defun interro (flux caractere) caractere                                     ;?
  (let ((sec (read-char flux))) (cond
    ((member sec '(#\_ #\$ #\% #\~)) (cond
        ((eq 'lpar (typechpeek))
         (append (list (concatenate 'string (string #\?) (string sec)))
                 (list nil)
                 (read flux)))
        ((not (eq 'pname (typechpeek)))
         (error "macro-caractere \"?\" : caractere errone, ou pas de ~
                 \"(\" apres \"~A\"." sec))
        (t (let ((nvar (read flux)))
             (append (list (concatenate 'string (string #\?) (string sec)))
                     (list nvar)
                     (if (not (eq 'lpar (typechpeek)))
                       (error "macro-caractere \"?\" : pas de \"(\" ~
                               apres la variable de filtre \"~A\"." nvar)
                       (read flux)))))))
    (t (error "macro-caractere \"?\" ne peut etre suivi de \"~A\"." sec)))))
(set-macro-character #\? #'interro)

(defun acommercial (flux caractere) caractere (cond                           ;@
    ((eq 'lpar (typechpeek))
     (append (list '|@|) (read flux)))
    ((not (eq 'pname (typechpeek)))
     (let ((car (read-char flux)))
       (error "macro-caractere \"@\" ne peut etre suivi de \"~A\" : ~
               doit etre suivi de \"(\" ou d'un nom de variable" car)))
    (t (cons '|@| (read flux)))))
(set-macro-character #\@ #'acommercial)

(defun circonflexe (flux caractere) caractere (cond                           ;^
    ((eq 'lpar (typechpeek))
     (append (list '|^|)
             (list nil)
             (read flux)))
    ((not (eq 'pname (typechpeek)))
     (let ((car (read-char flux)))
       (error "macro-caractere \"^\" ne peut etre suivi de \"~A\" : ~
               doit etre suivi de \"(\" ou d'un nom de variable" car)))
    (t (let ((nvar (read flux)))
         (append (list '|^|)
                 (list nvar)
                 (if (not (eq 'lpar (typechpeek)))
                   (error "macro-caractere \"^\" : pas de \"(\" ~
                           apres la variable de filtre \"~A\"." nvar)
                   (read flux)))))))
(set-macro-character #\^ #'circonflexe)

;;; Tableau de hachage common-lisp::typech : types des 128 caractres ASCII

;;   N.B. : Les divers caractres nomms #\control-<x> ne sont pas reconnus en
;; Macintosh Common LISP (MCL) ; ils ne sont d'ailleurs pas utiliss
;; spcifiquement pour les besoins des prsents filtres. Les types qui leur sont
;; attribus dans le tableau suivant correspondent  un environnement dvelopp
;; sous UNIX.

(setf (gethash    #\null             common-lisp::typech)    'null     )    ;000
;(setf (gethash    #\control-A        common-lisp::typech)    'macro    )   ;001
;(setf (gethash    #\control-B        common-lisp::typech)    'pname    )   ;002
;(setf (gethash    #\control-C        common-lisp::typech)    'ecom     )   ;003
;(setf (gethash    #\control-D        common-lisp::typech)    'ecom     )   ;004
;(setf (gethash    #\control-E        common-lisp::typech)    'macro    )   ;005
;(setf (gethash    #\control-F        common-lisp::typech)    'macro    )   ;006
(setf (gethash    #\bell             common-lisp::typech)    'pname    )    ;007
(setf (gethash    #\backspace        common-lisp::typech)    'sep      )    ;008
(setf (gethash    #\tab              common-lisp::typech)    'sep      )    ;009
(setf (gethash    #\linefeed         common-lisp::typech)    'ecom     )    ;010
;(setf (gethash    #\control-K        common-lisp::typech)    'sep      )   ;011
(setf (gethash    #\page             common-lisp::typech)    'ecom     )    ;012
(setf (gethash    #\return           common-lisp::typech)    'ecom     )    ;013
;(setf (gethash    #\control-N        common-lisp::typech)    'pname    )   ;014
;(setf (gethash    #\control-O        common-lisp::typech)    'pname    )   ;015
;(setf (gethash    #\control-P        common-lisp::typech)    'pname    )   ;016
;(setf (gethash    #\control-Q        common-lisp::typech)    'pname    )   ;017
;(setf (gethash    #\control-R        common-lisp::typech)    'pname    )   ;018
;(setf (gethash    #\control-S        common-lisp::typech)    'pname    )   ;019
;(setf (gethash    #\control-T        common-lisp::typech)    'pname    )   ;020
;(setf (gethash    #\control-U        common-lisp::typech)    'pname    )   ;021
;(setf (gethash    #\control-V        common-lisp::typech)    'pname    )   ;022
;(setf (gethash    #\control-W        common-lisp::typech)    'pname    )   ;023
;(setf (gethash    #\control-X        common-lisp::typech)    'pname    )   ;024
;(setf (gethash    #\control-Y        common-lisp::typech)    'pname    )   ;025
;(setf (gethash    #\control-Z        common-lisp::typech)    'pname    )   ;026
(setf (gethash    #\escape           common-lisp::typech)    'pname    )    ;027
;(setf (gethash    #\control-\        common-lisp::typech)    'pname    )   ;028
;(setf (gethash    #\control-]        common-lisp::typech)    'pname    )   ;029
;(setf (gethash    #\control-^        common-lisp::typech)    'pname    )   ;030
;(setf (gethash    #\control-_        common-lisp::typech)    'pname    )   ;031
(setf (gethash    #\space            common-lisp::typech)    'sep      )    ;032
(setf (gethash    #\!                common-lisp::typech)    'macro    )    ;033
(setf (gethash    #\"                common-lisp::typech)    'string   )    ;034
(setf (gethash    #\#                common-lisp::typech)    'splice   )    ;035
(setf (gethash    #\$                common-lisp::typech)    'filtre   )    ;036
(setf (gethash    #\%                common-lisp::typech)    'filtre   )    ;037
(setf (gethash    #\&                common-lisp::typech)    'filtre   )    ;038
(setf (gethash    #\'                common-lisp::typech)    'macro    )    ;039
(setf (gethash    #\(                common-lisp::typech)    'lpar     )    ;040
(setf (gethash    #\)                common-lisp::typech)    'rpar     )    ;041
(setf (gethash    #\*                common-lisp::typech)    'pname    )    ;042
(setf (gethash    #\+                common-lisp::typech)    'pname    )    ;043
(setf (gethash    #\,                common-lisp::typech)    'macro    )    ;044
(setf (gethash    #\-                common-lisp::typech)    'pname    )    ;045
(setf (gethash    #\.                common-lisp::typech)    'dot      )    ;046
(setf (gethash    #\/                common-lisp::typech)    'pname    )    ;047
(setf (gethash    #\0                common-lisp::typech)    'pname    )    ;048
(setf (gethash    #\1                common-lisp::typech)    'pname    )    ;049
(setf (gethash    #\2                common-lisp::typech)    'pname    )    ;050
(setf (gethash    #\3                common-lisp::typech)    'pname    )    ;051
(setf (gethash    #\4                common-lisp::typech)    'pname    )    ;052
(setf (gethash    #\5                common-lisp::typech)    'pname    )    ;053
(setf (gethash    #\6                common-lisp::typech)    'pname    )    ;054
(setf (gethash    #\7                common-lisp::typech)    'pname    )    ;055
(setf (gethash    #\8                common-lisp::typech)    'pname    )    ;056
(setf (gethash    #\9                common-lisp::typech)    'pname    )    ;057
(setf (gethash    #\:                common-lisp::typech)    'pname    )    ;058
(setf (gethash    #\;                common-lisp::typech)    'bcom     )    ;059
(setf (gethash    #\<                common-lisp::typech)    'pname    )    ;060
(setf (gethash    #\=                common-lisp::typech)    'pname    )    ;061
(setf (gethash    #\>                common-lisp::typech)    'pname    )    ;062
(setf (gethash    #\?                common-lisp::typech)    'filtre   )    ;063
(setf (gethash    #\@                common-lisp::typech)    'filtre   )    ;064
(setf (gethash    #\A                common-lisp::typech)    'pname    )    ;065
(setf (gethash    #\B                common-lisp::typech)    'pname    )    ;066
(setf (gethash    #\C                common-lisp::typech)    'pname    )    ;067
(setf (gethash    #\D                common-lisp::typech)    'pname    )    ;068
(setf (gethash    #\E                common-lisp::typech)    'pname    )    ;069
(setf (gethash    #\F                common-lisp::typech)    'pname    )    ;070
(setf (gethash    #\G                common-lisp::typech)    'pname    )    ;071
(setf (gethash    #\H                common-lisp::typech)    'pname    )    ;072
(setf (gethash    #\I                common-lisp::typech)    'pname    )    ;073
(setf (gethash    #\J                common-lisp::typech)    'pname    )    ;074
(setf (gethash    #\K                common-lisp::typech)    'pname    )    ;075
(setf (gethash    #\L                common-lisp::typech)    'pname    )    ;076
(setf (gethash    #\M                common-lisp::typech)    'pname    )    ;077
(setf (gethash    #\N                common-lisp::typech)    'pname    )    ;078
(setf (gethash    #\O                common-lisp::typech)    'pname    )    ;079
(setf (gethash    #\P                common-lisp::typech)    'pname    )    ;080
(setf (gethash    #\Q                common-lisp::typech)    'pname    )    ;081
(setf (gethash    #\R                common-lisp::typech)    'pname    )    ;082
(setf (gethash    #\S                common-lisp::typech)    'pname    )    ;083
(setf (gethash    #\T                common-lisp::typech)    'pname    )    ;084
(setf (gethash    #\U                common-lisp::typech)    'pname    )    ;085
(setf (gethash    #\V                common-lisp::typech)    'pname    )    ;086
(setf (gethash    #\W                common-lisp::typech)    'pname    )    ;087
(setf (gethash    #\X                common-lisp::typech)    'pname    )    ;088
(setf (gethash    #\Y                common-lisp::typech)    'pname    )    ;089
(setf (gethash    #\Z                common-lisp::typech)    'pname    )    ;090
(setf (gethash    #\[                common-lisp::typech)    'pname    )    ;091
(setf (gethash    #\\                common-lisp::typech)    'macro    )    ;092
(setf (gethash    #\]                common-lisp::typech)    'pname    )    ;093
(setf (gethash    #\^                common-lisp::typech)    'filtre   )    ;094
(setf (gethash    #\_                common-lisp::typech)    'filtre   )    ;095
(setf (gethash    #\`                common-lisp::typech)    'macro    )    ;096
(setf (gethash    #\a                common-lisp::typech)    'pname    )    ;097
(setf (gethash    #\b                common-lisp::typech)    'pname    )    ;098
(setf (gethash    #\c                common-lisp::typech)    'pname    )    ;099
(setf (gethash    #\d                common-lisp::typech)    'pname    )    ;100
(setf (gethash    #\e                common-lisp::typech)    'pname    )    ;101
(setf (gethash    #\f                common-lisp::typech)    'pname    )    ;102
(setf (gethash    #\g                common-lisp::typech)    'pname    )    ;103
(setf (gethash    #\h                common-lisp::typech)    'pname    )    ;104
(setf (gethash    #\i                common-lisp::typech)    'pname    )    ;105
(setf (gethash    #\j                common-lisp::typech)    'pname    )    ;106
(setf (gethash    #\k                common-lisp::typech)    'pname    )    ;107
(setf (gethash    #\l                common-lisp::typech)    'pname    )    ;108
(setf (gethash    #\m                common-lisp::typech)    'pname    )    ;109
(setf (gethash    #\n                common-lisp::typech)    'pname    )    ;110
(setf (gethash    #\o                common-lisp::typech)    'pname    )    ;111
(setf (gethash    #\p                common-lisp::typech)    'pname    )    ;112
(setf (gethash    #\q                common-lisp::typech)    'pname    )    ;113
(setf (gethash    #\r                common-lisp::typech)    'pname    )    ;114
(setf (gethash    #\s                common-lisp::typech)    'pname    )    ;115
(setf (gethash    #\t                common-lisp::typech)    'pname    )    ;116
(setf (gethash    #\u                common-lisp::typech)    'pname    )    ;117
(setf (gethash    #\v                common-lisp::typech)    'pname    )    ;118
(setf (gethash    #\w                common-lisp::typech)    'pname    )    ;119
(setf (gethash    #\x                common-lisp::typech)    'pname    )    ;120
(setf (gethash    #\y                common-lisp::typech)    'pname    )    ;121
(setf (gethash    #\z                common-lisp::typech)    'pname    )    ;122
(setf (gethash    #\{                common-lisp::typech)    'pname    )    ;123
(setf (gethash    #\|                common-lisp::typech)    'macro    )    ;124
(setf (gethash    #\}                common-lisp::typech)    'pname    )    ;125
(setf (gethash    #\~                common-lisp::typech)    'filtre   )    ;126
(setf (gethash    #\delete           common-lisp::typech)    'null     )    ;127
